// file:    array.H
// author:  Robert M. Keller
// purpose: Provide class definition for extendible array
// $Id: array.H,v 1.3 1995/11/27 19:33:32 keller Exp keller $

//
// array<T> class (where T is a previously-defined type)
//
// usage: array<T> A(long length); constructs array

//        A[i] gives i-th element

//        A.length() returns length

#ifndef __ARRAY_H__

#include <iostream>
#include <stdlib.h>

template <class T>
class array
{
protected:

long Length;                               // length of the array
T *a;                                      // the actual array

public: 

array();                                   // default constructor (length 0 array)

array(long length);                        // length-specifying constructor

array(array<T> &original);                 // copy constructor

inline T& operator[](long i) const;        // access element i

inline long length() const;                // return current length

inline array<T> & length(long new_length); // resize the array

inline array<T> & extend(long inc);        // extend the array by increment

array<T> & operator=(array<T> & source);   // assignment operator

~array();                                  // destructor

private:

void construct(long Length);               // called by constructors

void copy(T* b, T* a, long limit);         // copy contents        
};  // class array


//
// default constructor
//

template <class T>
array<T>::array()                       
{ 
construct(0); 
}

//
// constructor given length
//

template <class T>
array<T>::array(long Length)
{ 
construct(Length); 
}


//
// copy constructor
//

template <class T>
array<T>::array(array<T> &original)
{
construct(original.length());
copy(a, original.a, Length);
}


//
// return length
//

template <class T>
inline long array<T>::length() const
{ 
return Length; 
}               


//
// return ref. to i-th element
//

template <class T>
inline T& array<T>::operator[](long i) const
{
return a[i];
}       


//
// resize the array
//

template <class T>
array<T>& array<T>::length(long new_length)
{
T* new_array = new T[new_length];       // get new space
if( !new_array )                
  {
  std::cerr << "not enough space to resize array to " << new_length << std::endl;
  exit(1);
  }

// copy smaller number of elements, length vs. new_length
copy(new_array, a, (new_length < Length ? new_length : Length));

delete [] a;                            // delete old space
a = new_array;                          // install new
Length = new_length;                    // set new length
return *this;
}


//
// extend the array by specified increment inc
//

template <class T>
inline array<T> & array<T>::extend(long inc)
{
length( length()+inc );
return *this;
}


//
// assignment operator
//

template <class T>
array<T>& array<T>::operator=(array<T> & source)
{
if( this == &source )
  return *this;                         // bypass self-assignment
if( Length != source.length() )         // bypass reallocation if length same
  length(source.length());              // reallocate and set target length
copy(a, source.a, Length);              // copy elements
return *this;                           // return target
}


//
// construct is used by constructors to initialize and allocate space
//

template <class T>
void array<T>::construct(long length)
{
Length = length;
a = new T[Length];
if( !a )
  {
  std::cerr << "not enough space to allocate array of length " << length << std::endl;
  exit(1);
  }
}


//
// copy array contents from a to b
//

template <class T>
void array<T>::copy(T* b, T* a, long limit)
{
for( long i = 0; i < limit; i++ )       // copy elements
  b[i] = a[i];
}


//
// destructor
//

template <class T>
array<T>::~array()
{
delete [] a;
}

/*
template <class T>
ostream & operator<<(ostream & out, array<T> & a)
{
for( long i = 0; i < a.length(); i++ )
  out << a[i] << " ";
return out;
}
*/

#define __ARRAY_H__
#endif

