// file:    symbol_table.h
// author:  R. Keller
// purpose: Implement simple symbol table
/*
The symbol_table type implemented here as a class is a special type
of partial function between strings and ints.

A partial function is a set of source-target pairs, such that for
a given source there is at most one target.  A symbol table is
a combination of a partial function from sources to targets, along
with a set of sources that includes (but does not necessarily equal)
the sources with which a target is associated.

If a source is in the set, but there is no corresponding target,
we will say the source is 'undefined'.  If the source is not
in the set, we say the source is 'unknown'.  

Thus, for a given source, there are four exclusive possibilities:
 _ The source has a unique associated target.
 _ The source is undefined.
 _ The source is unknown.
 _ The source has multiple definitions (only the last target is retained)

Example of use:
In processing a program's text, the source corresponds to an identifier
and the target the meaning (such as a storage location) for that identifier.
If the first encounter of a source is the definition of the source,
then its target is immediately known and the source is entered along with
the target into the symbol table, with definitions set to 1.  On the other 
hand, if the first encounter of a source is a use, then the source is entered 
with definitions set to 0, without a target.

A source is regarded as unknown if it has not been seen before
as a definition or use.  Note that there may be multiple encounters of
a source as definitions.  This could represent an error, depending on the
whether the language allows redefinition or not.  The present class remembers
the number of definitions and only the last associated value.

The following are the principal procedures:

 symbol_table(int inc) constructs a symbol table with initial size, inc.

 find_source(char* source) returns the index of the source, or -1 if no index.

 new_source(char* source) adds a source as known, but undefined, returning the 
   index.  This does not check that the source is actually new.  That should be 
   done first using find_source.

 ensure_source(char* source) makes sure the source is in the table,
   using new_source to add it if not.  The index is returned.

 set_target(int index, int target) sets the target associated with the 
   source at index.

 defined(int index) indicates the number of definitions of the source at index

 get_target(int index) returns the target associated with the source.
   Note that if defined(index) is 0, then this value will be garbage.

 add(std::string source, int target) adds source to table as if a new source,
   and associates target with it.  Note that this does not check for
   presence of the source already.  If that should be the case, then
   find will not work properly.  If in doubt, use update instead, or use a 
   combination of the other primitives.

 update(std::string source, int target) ensures that source is in the table, then
   sets the corresponding target to the given argument

 number() returns the number of sources in the table

 clear()  clears the table

 extend(int inc) increases table space by inc.  This is called automatically
   by new_source, etc.

Except for clear(), this class does not allow provide for removal of 
source-target pairs.

The absolute position of pairs is maintained, so one can additionally 
use the index to advantage to index other arrays.  The index can also
be used to list or transfer the entire contents of the partial function.
*/

#include "newstr.cc"                // to get newstr, which copies string

typedef struct                      // define pair struct, stored in bag
{
  std::string source;                     // source part of a mapped pair
  int target;                      // target part of a mapped pair
  int definitions;                 // indicates whether pending
} pair;  

class symbol_table
{
public: 

symbol_table(int inc);               // constructor

void clear();                             // empty the table

int new_source(std::string source);            // add source, returning index

int find_source(std::string source);           // find source, returning index

int ensure_source(std::string source);         // find, or add if necessary

void set_target(int index, int target); // set target at index

int get_target(int index);              // get target at index

void undefine(int index);                // make source undefined

std::string get_source(int index);             // get source at index

int defined(int index);                  // return number of definitions

void add(std::string source, int target);      // add new source-target pair

void update(std::string source, int target);   // change or add source-target pair

void extend(int inc);                    // increase space

int number();                            // return number of entries in table

private:

pair *pairs;                       // array of std::string for pairs
int num_pairs;                    // number of pairs
int space;                        // amount of space in array
int increment;                    // incremental amount of space
};

// constructor for symbol_table; 'inc' specifies initial size and
// automatic increment to be used when extending

symbol_table:: symbol_table(int inc)
{
increment = inc;                   // remember the increment
pairs = new pair [increment];      // allocate space
space = increment;                 // remember how much
clear();                           // finish initializing
}

// reset the symbol_table to empty, retaining space

void symbol_table:: clear()
{
num_pairs = 0;
}

// return the number of pairs currently in the symbol_table

int symbol_table:: number()
{
return num_pairs;
}

// enter a new source, target undefined

int symbol_table:: new_source(std::string source)
{
int index;
if( num_pairs == space )
  extend(increment);
index = num_pairs;
num_pairs++;                                   // one more pair
pairs[index].source = source;          // copy source in
pairs[index].definitions = 0;                  // set pending
return index;
}

// find the source, giving its index, or return -1 if not present

int symbol_table:: find_source(std::string source)
{
for( int index = 0; index < num_pairs; index++ )
  if( pairs[index].source == source )
    return index;                              // found source
return -1;                                     // source not found
}

// ensure the source is in the table, returning its index
int symbol_table:: ensure_source(std::string source)
{
int index = find_source(source);
if( index >= 0 )
  return index;
return new_source(source);
}

// set the target at index, indicating number of definitions

void symbol_table:: set_target(int index, int target)
{
pairs[index].target = target;
pairs[index].definitions++;
}

// make undefined

void symbol_table:: undefine(int index)
{
pairs[index].definitions = 0;
}

// indicate number of definitions of source

int symbol_table:: defined(int index)
{
return pairs[index].definitions;
}

// return target at index

int symbol_table:: get_target(int index)
{
return pairs[index].target;
}

// return source at index

std::string symbol_table:: get_source(int index)
{
return pairs[index].source;
}

// change or add source-target pair

void symbol_table:: update(std::string source, int new_target)
{
set_target(ensure_source(source), new_target);
}

// add source-target pair

void symbol_table:: add(std::string source, int new_target)
{
set_target(new_source(source), new_target);
}

// extend the table by inc pairs

void symbol_table:: extend(int inc)        
{
int new_space = space + inc;           // compute new space amount
pair *new_pairs = new pair[new_space];  // allocate new space

for( int i = 0; i < num_pairs; i++ )   // copy pairs to new array
  new_pairs[i] = pairs[i];

delete [] pairs;                        // delete old array
pairs = new_pairs;                      // establish new array
space = new_space;                      // establish new space
}

